@@ -27,14 +27,10 @@
 #include <config.h>
 
 #include        "gsignal.h"
-
 #include        "gbsearcharray.h"
 #include        "gvaluecollector.h"
 #include	"gvaluetypes.h"
 #include	"gboxed.h"
-
-#include	"gobjectalias.h"
-
 #include	<string.h> 
 #include	<signal.h>
 
@@ -44,7 +40,7 @@
 #define	MAX_STACK_VALUES	(16)
 #define HANDLER_PRE_ALLOC       (48)
 
-#define REPORT_BUG      "please report occurrence circumstances to gtk-devel-list@gnome.org"
+#define REPORT_BUG      "please report occourance circumstances to gtk-devel-list@gnome.org"
 #ifdef	G_ENABLE_DEBUG
 #define IF_DEBUG(debug_type, cond)	if ((_g_type_debug_flags & G_TYPE_DEBUG_ ## debug_type) || cond)
 static volatile gpointer g_trace_instance_signals = NULL;
@@ -212,10 +208,7 @@ struct _HandlerList
 {
   guint    signal_id;
   Handler *handlers;
-  Handler *tail_before;  /* normal signal handlers are appended here  */
-  Handler *tail_after;   /* CONNECT_AFTER handlers are appended here  */
 };
-
 struct _Handler
 {
   gulong        sequential_number;
@@ -300,7 +293,7 @@ signal_id_lookup (GQuark quark,
 
   key.quark = quark;
 
-  /* try looking up signals for this type and its ancestors */
+  /* try looking up signals for this type and its anchestors */
   do
     {
       SignalKey *signal_key;
@@ -361,9 +354,7 @@ handler_list_ensure (guint    signal_id,
   HandlerList key;
   
   key.signal_id = signal_id;
-  key.handlers    = NULL;
-  key.tail_before = NULL;
-  key.tail_after  = NULL;
+  key.handlers = NULL;
   if (!hlbsa)
     {
       hlbsa = g_bsearch_array_create (&g_signal_hlbsa_bconfig);
@@ -585,45 +576,16 @@ handler_unref_R (guint    signal_id,
   handler->ref_count -= 1;
   if (!handler->ref_count)
     {
-      HandlerList *hlist = NULL;
-
       if (handler->next)
         handler->next->prev = handler->prev;
-      if (handler->prev)    /* watch out for g_signal_handlers_destroy()! */
+      if (handler->prev)	/* watch out for g_signal_handlers_destroy()! */
         handler->prev->next = handler->next;
       else
         {
-          hlist = handler_list_lookup (signal_id, instance);
+          HandlerList *hlist = handler_list_lookup (signal_id, instance);
+          
           hlist->handlers = handler->next;
         }
-
-      if (instance)
-        {
-          /*  check if we are removing the handler pointed to by tail_before  */
-          if (!handler->after && (!handler->next || handler->next->after))
-            {
-              if (!hlist)
-                hlist = handler_list_lookup (signal_id, instance);
-              if (hlist)
-                {
-                  g_assert (hlist->tail_before == handler); /* paranoid */
-                  hlist->tail_before = handler->prev;
-                }
-            }
-
-          /*  check if we are removing the handler pointed to by tail_after  */
-          if (!handler->next)
-            {
-              if (!hlist)
-                hlist = handler_list_lookup (signal_id, instance);
-              if (hlist)
-                {
-                  g_assert (hlist->tail_after == handler); /* paranoid */
-                  hlist->tail_after = handler->prev;
-                }
-            }
-        }
-
       SIGNAL_UNLOCK ();
       g_closure_unref (handler->closure);
       SIGNAL_LOCK ();
@@ -642,38 +604,29 @@ handler_insert (guint    signal_id,
   
   hlist = handler_list_ensure (signal_id, instance);
   if (!hlist->handlers)
+    hlist->handlers = handler;
+  else if (hlist->handlers->after && !handler->after)
     {
+      handler->next = hlist->handlers;
+      hlist->handlers->prev = handler;
       hlist->handlers = handler;
-      if (!handler->after)
-        hlist->tail_before = handler;
-    }
-  else if (handler->after)
-    {
-      handler->prev = hlist->tail_after;
-      hlist->tail_after->next = handler;
     }
   else
     {
-      if (hlist->tail_before)
-        {
-          handler->next = hlist->tail_before->next;
-          if (handler->next)
-            handler->next->prev = handler;
-          handler->prev = hlist->tail_before;
-          hlist->tail_before->next = handler;
-        }
-      else /* insert !after handler into a list of only after handlers */
-        {
-          handler->next = hlist->handlers;
-          if (handler->next)
-            handler->next->prev = handler;
-          hlist->handlers = handler;
-        }
-      hlist->tail_before = handler;
+      Handler *tmp = hlist->handlers;
+      
+      if (handler->after)
+        while (tmp->next)
+          tmp = tmp->next;
+      else
+        while (tmp->next && !tmp->next->after)
+          tmp = tmp->next;
+      if (tmp->next)
+        tmp->next->prev = handler;
+      handler->next = tmp->next;
+      handler->prev = tmp;
+      tmp->next = handler;
     }
-
-  if (!handler->next)
-    hlist->tail_after = handler;
 }
 
 static inline void
@@ -756,7 +709,7 @@ signal_key_cmp (gconstpointer node1,
 }
 
 void
-g_signal_init (void)
+g_signal_init (void) /* sync with gtype.c */
 {
   SIGNAL_LOCK ();
   if (!g_n_signal_nodes)
@@ -871,18 +824,12 @@ g_signal_add_emission_hook (guint               signal_id,
 
   SIGNAL_LOCK ();
   node = LOOKUP_SIGNAL_NODE (signal_id);
-  if (!node || node->destroyed)
+  if (!node || node->destroyed || (node->flags & G_SIGNAL_NO_HOOKS))
     {
       g_warning ("%s: invalid signal id `%u'", G_STRLOC, signal_id);
       SIGNAL_UNLOCK ();
       return 0;
     }
-  if (node->flags & G_SIGNAL_NO_HOOKS) 
-    {
-      g_warning ("%s: signal id `%u' does not support emission hooks (G_SIGNAL_NO_HOOKS flag set)", G_STRLOC, signal_id);
-      SIGNAL_UNLOCK ();
-      return 0;
-    }
   if (detail && !(node->flags & G_SIGNAL_DETAILED))
     {
       g_warning ("%s: signal id `%u' does not support detail (%u)", G_STRLOC, signal_id, detail);
@@ -1800,6 +1747,7 @@ g_signal_handler_is_connected (gpointer instance,
   gboolean connected;
 
   g_return_val_if_fail (G_TYPE_CHECK_INSTANCE (instance), FALSE);
+  g_return_val_if_fail (handler_id > 0, FALSE);
 
   SIGNAL_LOCK ();
   handler = handler_lookup (instance, handler_id, NULL);
@@ -2268,7 +2216,7 @@ g_signal_emit_valist (gpointer instance,
 	  g_free (error);
 	  
 	  /* we purposely leak the value here, it might not be
-	   * in a sane state if an error condition occured
+	   * in a sane state if an error condition occoured
 	   */
 	}
     }
@@ -2658,6 +2606,3 @@ g_signal_accumulator_true_handled (GSignalInvocationHint *ihint,
 #include	"gobject.h"
 #include	"genums.h"
 #include        "gmarshal.c"
-
-#define __G_SIGNAL_C__
-#include "gobjectaliasdef.c"
